﻿Public Class HierarchyNavigate1


    Private Sub SetInputMapRecursive(inputMap As FarPoint.Win.Spread.InputMap, root As FarPoint.Win.Spread.SpreadView, spreadView As FarPoint.Win.Spread.SpreadView)
        spreadView.SetInputMap(FarPoint.Win.Spread.InputMapMode.WhenFocused, inputMap)
        spreadView.GetActionMap().Put("HierarchyMoveToPreviousRow", New HierarchyMoveToPreviousRowAction())
        spreadView.GetActionMap().Put("HierarchyMoveToNextRow", New HierarchyMoveToNextRowAction())
        System.Diagnostics.Debug.WriteLine("Setting input map to view")
        inputMap.Put(New FarPoint.Win.Spread.Keystroke(Keys.Up, Keys.None), "HierarchyMoveToPreviousRow")
        inputMap.Put(New FarPoint.Win.Spread.Keystroke(Keys.Down, Keys.None), "HierarchyMoveToNextRow")
        Dim children As ArrayList = spreadView.GetChildWorkbooks()
        For Each child As FarPoint.Win.Spread.SpreadView In children
            SetInputMapRecursive(inputMap, root, child)
        Next

    End Sub

    Private Sub ExpandChildrenRecursive(sv As FarPoint.Win.Spread.SheetView)
        sv.ExpandRow(-1, True)
        For Each child As FarPoint.Win.Spread.SheetView In sv.GetChildSheets()
            ExpandChildrenRecursive(child)
        Next
    End Sub

    Private Function CreateDataSource() As DataSet
        Dim ds As New DataSet()
        Dim dtOffices As DataTable = ds.Tables.Add("Offices")
        Dim dtPeople As DataTable = ds.Tables.Add("People")
        Dim dtCars As DataTable = ds.Tables.Add("Cars")
        Dim dtPets As DataTable = ds.Tables.Add("Pets")

        dtOffices.Columns.AddRange(New DataColumn() {New DataColumn("Location"), New DataColumn("OfficeID", GetType(Integer))})
        dtOffices.Rows.Add(New Object() {"Raleigh, NC", 0})
        dtOffices.Rows.Add(New Object() {"Pittsburgh, PA", 1})
        dtOffices.Rows.Add(New Object() {"Kirkland, WA", 2})

        dtPeople.Columns.AddRange(New DataColumn() {New DataColumn("FirstName"), New DataColumn("LastName"), New DataColumn("PersonID", GetType(Integer)), New DataColumn("OfficeID", GetType(Integer))})
        dtPeople.Rows.Add(New Object() {"Sean", "Fieldes", 0, 0})
        dtPeople.Rows.Add(New Object() {"Dan", "Smith", 1, 0})
        dtPeople.Rows.Add(New Object() {"Bo", "Jones", 2, 0})
        dtPeople.Rows.Add(New Object() {"Kevin", "Hill", 3, 0})
        dtPeople.Rows.Add(New Object() {"Gina", "West", 4, 0})
        dtPeople.Rows.Add(New Object() {"John", "Young", 5, 1})
        dtPeople.Rows.Add(New Object() {"Andrew", "Perkins", 6, 1})
        dtPeople.Rows.Add(New Object() {"Chris", "Davis", 7, 1})
        dtPeople.Rows.Add(New Object() {"Eric", "Johnson", 8, 1})
        dtPeople.Rows.Add(New Object() {"Justin", "Thomas", 9, 1})
        dtPeople.Rows.Add(New Object() {"Devinder", "Lee", 10, 2})
        dtPeople.Rows.Add(New Object() {"Sanjeev", "Clark", 11, 2})
        dtPeople.Rows.Add(New Object() {"Alex", "Walker", 12, 2})

        dtCars.Columns.AddRange(New DataColumn() {New DataColumn("Make"), New DataColumn("Model"), New DataColumn("Color"), New DataColumn("PersonID", GetType(Integer))})
        dtCars.Rows.Add(New Object() {"Honda", "CR-V", "Silver", 0})
        dtCars.Rows.Add(New Object() {"Saab", "97-X", "Black", 0})
        dtCars.Rows.Add(New Object() {"Jeep", "Cherokee", "Black", 1})
        dtCars.Rows.Add(New Object() {"Ford", "Mustang", "Black", 2})
        dtCars.Rows.Add(New Object() {"Volkswagon", "Beetle", "Black", 3})
        dtCars.Rows.Add(New Object() {"Honda", "CR-V", "Blue", 4})
        dtCars.Rows.Add(New Object() {"Ferari", "Testarosa", "Red", 4})
        dtCars.Rows.Add(New Object() {"Ford", "F150", "Red", 5})
        dtCars.Rows.Add(New Object() {"Honda", "Accord", "Silver", 6})
        dtCars.Rows.Add(New Object() {"Toyota", "Tundra", "Black", 7})
        dtCars.Rows.Add(New Object() {"Honda", "Odyssey", "Silver", 8})
        dtCars.Rows.Add(New Object() {"Jeep", "Wrangler", "Silver", 9})
        dtCars.Rows.Add(New Object() {"Mazarati", "Quatroport", "Black", 10})
        dtCars.Rows.Add(New Object() {"Lexus", "LX", "Black", 11})
        dtCars.Rows.Add(New Object() {"Honda", "Accord", "Black", 12})

        dtPets.Columns.AddRange(New DataColumn() {New DataColumn("Name"), New DataColumn("Type"), New DataColumn("Color"), New DataColumn("Gender"), New DataColumn("PersonID", GetType(Integer))})
        dtPets.Rows.Add(New Object() {"Putzel", "Cat", "Orange", "Male", 0})
        dtPets.Rows.Add(New Object() {"Hickory", "Cat", "Black", "Male", 0})
        dtPets.Rows.Add(New Object() {"Ginger", "Cat", "Orange", "Female", 0})
        dtPets.Rows.Add(New Object() {"Spice", "Cat", "Orange", "Male", 0})
        dtPets.Rows.Add(New Object() {"Cupcake", "Cat", "Gray", "Male", 0})
        dtPets.Rows.Add(New Object() {"Kitkat", "Cat", "Gray", "Male", 0})
        dtPets.Rows.Add(New Object() {"Sweetie", "Cat", "Black/White Tabby", "Female", 0})
        dtPets.Rows.Add(New Object() {"Fido", "Dog", "Black", "Male", 1})
        dtPets.Rows.Add(New Object() {"Herbie", "Hamster", "Brown", 2})
        dtPets.Rows.Add(New Object() {"Ricky", "Rhinoseros", "Black", 3})
        dtPets.Rows.Add(New Object() {"Freddy", "Penguin", "Black and White", 4})
        dtPets.Rows.Add(New Object() {"Kermit", "Frog", "Green", 5})
        dtPets.Rows.Add(New Object() {"Fozzy", "Bear", "Brown", 6})
        dtPets.Rows.Add(New Object() {"Tim", "Turtle", "Green", 7})
        dtPets.Rows.Add(New Object() {"Wanda", "Fish", "Rainbow", 8})
        dtPets.Rows.Add(New Object() {"Bruce", "Dog", "Brown", 9})

        ds.Relations.Add(New DataRelation("OfficePerson", dtOffices.Columns("OfficeID"), dtPeople.Columns("OfficeID")))
        ds.Relations.Add(New DataRelation("PersonCar", dtPeople.Columns("PersonID"), dtCars.Columns("PersonID")))
        ds.Relations.Add(New DataRelation("PersonPet", dtPeople.Columns("PersonID"), dtPets.Columns("PersonID")))
        Return ds
    End Function


    Private Class HierarchyMoveToPreviousRowAction
        Inherits FarPoint.Win.Spread.Action
        Private Function FindLastChildRecursive(spreadView As FarPoint.Win.Spread.SpreadView) As FarPoint.Win.Spread.SpreadView
            Dim children As ArrayList = spreadView.GetChildWorkbooks()
            If children Is Nothing OrElse children.Count = 0 Then
                Return spreadView
            Else
                For i As Integer = children.Count - 1 To 0 Step -1
                    Dim sheet As FarPoint.Win.Spread.SheetView = DirectCast(children(i), FarPoint.Win.Spread.SpreadView).GetSheetView()
                    If sheet.RowCount > 0 Then
                        Return FindLastChildRecursive(DirectCast(children(i), FarPoint.Win.Spread.SpreadView))
                    End If
                Next
            End If
            Return spreadView
        End Function
        Public Overrides Sub PerformAction(sender As Object)
            Dim spreadView As FarPoint.Win.Spread.SpreadView = TryCast(sender, FarPoint.Win.Spread.SpreadView)
            If spreadView IsNot Nothing Then
                Dim sheetView As FarPoint.Win.Spread.SheetView = spreadView.GetSheetView()
                Dim activeColumn As Integer = sheetView.ActiveColumnIndex
                Dim activeRow As Integer = sheetView.ActiveRowIndex
                Dim parentRowIndex As Integer = sheetView.ParentRowIndex
                Dim parentSpreadView As FarPoint.Win.Spread.SpreadView = spreadView.Parent
                If parentSpreadView IsNot Nothing Then
                    ' spreadView is a child workbook
                    Dim siblings As ArrayList = parentSpreadView.GetChildWorkbooks()
                    Dim children As ArrayList = spreadView.GetChildWorkbooks()
                    Dim childViewIndex As Integer = siblings.IndexOf(spreadView)
                    If activeRow = 0 AndAlso childViewIndex > 0 AndAlso DirectCast(siblings(childViewIndex - 1), FarPoint.Win.Spread.SpreadView).GetSheetView().ParentRowIndex = parentRowIndex Then
                        ' there is a sibling sheet above this one, go to its last row
                        Dim targetSpreadView As FarPoint.Win.Spread.SpreadView = DirectCast(siblings(childViewIndex - 1), FarPoint.Win.Spread.SpreadView)
                        Dim targetSheetView As FarPoint.Win.Spread.SheetView = targetSpreadView.GetSheetView()
                        spreadView.GetRootWorkbook().SetActiveWorkbook(targetSpreadView)
                        targetSheetView.SetActiveCell(targetSheetView.RowCount - 1, activeColumn)
                        targetSpreadView.ShowCell(targetSpreadView.GetActiveRowViewportIndex(), targetSpreadView.GetActiveColumnViewportIndex(), targetSheetView.ActiveRowIndex, targetSheetView.ActiveColumnIndex, FarPoint.Win.Spread.VerticalPosition.Nearest, FarPoint.Win.Spread.HorizontalPosition.Nearest)
                        Return
                    Else
                        Dim targetSheetView As FarPoint.Win.Spread.SheetView = Nothing
                        If activeRow > 0 AndAlso sheetView.ChildRelationCount > 0 Then
                            ' there are child sheets for the previous row, go to the last child's last row (unless they're all empty)
                            For n As Integer = sheetView.ChildRelationCount - 1 To 0 Step -1
                                targetSheetView = sheetView.GetChildView(activeRow - 1, n)
                                If targetSheetView.RowCount > 0 Then
                                    Exit For
                                End If
                            Next
                            If targetSheetView.RowCount > 0 Then
                                Dim targetSpreadView As FarPoint.Win.Spread.SpreadView = targetSheetView.ContainingViews(0)
                                spreadView.GetRootWorkbook().SetActiveWorkbook(targetSpreadView)
                                targetSheetView.SetActiveCell(targetSheetView.RowCount - 1, targetSheetView.ActiveColumnIndex)
                                targetSpreadView.ShowCell(targetSpreadView.GetActiveRowViewportIndex(), targetSpreadView.GetActiveColumnViewportIndex(), targetSheetView.ActiveRowIndex, targetSheetView.ActiveColumnIndex, FarPoint.Win.Spread.VerticalPosition.Nearest, FarPoint.Win.Spread.HorizontalPosition.Nearest)
                                Return
                            End If
                        End If
                        If activeRow = 0 Then
                            ' there is no sibling sheet above this one, and no child sheets for the previous row, go to the parent row
                            spreadView.GetRootWorkbook().SetActiveWorkbook(parentSpreadView)
                            targetSheetView = parentSpreadView.GetSheetView()
                            targetSheetView.SetActiveCell(sheetView.ParentRowIndex, targetSheetView.ActiveColumnIndex)
                            parentSpreadView.ShowCell(parentSpreadView.GetActiveRowViewportIndex(), parentSpreadView.GetActiveColumnViewportIndex(), targetSheetView.ActiveRowIndex, targetSheetView.ActiveColumnIndex, FarPoint.Win.Spread.VerticalPosition.Nearest, FarPoint.Win.Spread.HorizontalPosition.Nearest)
                            Return
                        End If
                    End If
                Else
                    ' spreadView is top-level view (root workbook)
                    Dim children As ArrayList = spreadView.GetChildWorkbooks()
                    For Each child As FarPoint.Win.Spread.SpreadView In children
                        If child.GetSheetView().ParentRowIndex = activeRow - 1 Then
                            Dim targetSpreadView As FarPoint.Win.Spread.SpreadView = FindLastChildRecursive(child)
                            Dim targetSheetView As FarPoint.Win.Spread.SheetView = targetSpreadView.GetSheetView()
                            spreadView.GetRootWorkbook().SetActiveWorkbook(targetSpreadView)
                            targetSheetView.SetActiveCell(targetSheetView.RowCount - 1, targetSheetView.ActiveColumnIndex)
                            targetSpreadView.ShowCell(targetSpreadView.GetActiveRowViewportIndex(), targetSpreadView.GetActiveColumnViewportIndex(), targetSheetView.ActiveRowIndex, targetSheetView.ActiveColumnIndex, FarPoint.Win.Spread.VerticalPosition.Nearest, FarPoint.Win.Spread.HorizontalPosition.Nearest)
                            Return
                        End If
                    Next
                End If
                ' do default action
                spreadView.GetRootWorkbook().GetActionMap().[Get]("MoveToPreviousRow").PerformAction(sender)
            End If
        End Sub
    End Class


    Private Class HierarchyMoveToNextRowAction
        Inherits FarPoint.Win.Spread.Action

        Private Function FindFirstChildRecursive(spreadView As FarPoint.Win.Spread.SpreadView) As FarPoint.Win.Spread.SpreadView
            Dim children As ArrayList = spreadView.GetChildWorkbooks()
            If children Is Nothing OrElse children.Count = 0 Then
                Return spreadView
            Else
                For i As Integer = 0 To children.Count - 2
                    Dim sheet As FarPoint.Win.Spread.SheetView = DirectCast(children(i), FarPoint.Win.Spread.SpreadView).GetSheetView()
                    If sheet.RowCount > 0 Then
                        Return FindFirstChildRecursive(DirectCast(children(i), FarPoint.Win.Spread.SpreadView))
                    End If
                Next
            End If
            Return spreadView
        End Function

        Public Overrides Sub PerformAction(sender As Object)
            Dim spreadView As FarPoint.Win.Spread.SpreadView = TryCast(sender, FarPoint.Win.Spread.SpreadView)
            If spreadView IsNot Nothing Then
                Dim sheetView As FarPoint.Win.Spread.SheetView = spreadView.GetSheetView()
                Dim activeColumn As Integer = sheetView.ActiveColumnIndex
                Dim activeRow As Integer = sheetView.ActiveRowIndex
                Dim parentRowIndex As Integer = sheetView.ParentRowIndex
                Dim parentSpreadView As FarPoint.Win.Spread.SpreadView = spreadView.Parent
                If parentSpreadView IsNot Nothing Then
                    ' spreadView is a child workbook
                    Dim targetSpreadView As FarPoint.Win.Spread.SpreadView = Nothing
                    Dim targetSheetView As FarPoint.Win.Spread.SheetView = Nothing
                    Dim siblings As ArrayList = parentSpreadView.GetChildWorkbooks()
                    Dim children As ArrayList = spreadView.GetChildWorkbooks()
                    Dim childViewIndex As Integer = siblings.IndexOf(spreadView)

                    If activeRow = sheetView.RowCount - 1 AndAlso childViewIndex < siblings.Count - 1 AndAlso DirectCast(siblings(childViewIndex + 1), FarPoint.Win.Spread.SpreadView).GetSheetView().ParentRowIndex = parentRowIndex Then
                        ' there is a sibling sheet below this one, go to its first row (unless they're all empty)
                        Dim n As Integer = childViewIndex + 1
                        While n < siblings.Count AndAlso DirectCast(siblings(n), FarPoint.Win.Spread.SpreadView).GetSheetView().ParentRowIndex = parentRowIndex
                            targetSpreadView = DirectCast(siblings(n), FarPoint.Win.Spread.SpreadView)
                            targetSheetView = targetSpreadView.GetSheetView()
                            If targetSheetView.RowCount > 0 Then
                                Exit While
                            End If
                            n += 1
                        End While
                        If targetSheetView.RowCount > 0 Then
                            spreadView.GetRootWorkbook().SetActiveWorkbook(targetSpreadView)
                            targetSheetView.SetActiveCell(0, activeColumn)
                            targetSpreadView.ShowCell(targetSpreadView.GetActiveRowViewportIndex(), targetSpreadView.GetActiveColumnViewportIndex(), targetSheetView.ActiveRowIndex, targetSheetView.ActiveColumnIndex, FarPoint.Win.Spread.VerticalPosition.Nearest, FarPoint.Win.Spread.HorizontalPosition.Nearest)
                            targetSheetView.Cells(targetSheetView.ActiveRowIndex, targetSheetView.ActiveColumnIndex).Invalidate()
                            Return
                        End If
                    End If

                    If sheetView.ChildRelationCount > 0 Then
                        ' there are child sheets for the next row, go to the first child's first row (unless they're all empty)
                        For n As Integer = 0 To sheetView.ChildRelationCount - 2
                            targetSheetView = sheetView.GetChildView(activeRow, n)
                            If targetSheetView.RowCount > 0 Then
                                Exit For
                            End If
                        Next
                        If targetSheetView.RowCount > 0 Then
                            targetSpreadView = targetSheetView.ContainingViews(0)
                            spreadView.GetRootWorkbook().SetActiveWorkbook(targetSpreadView)
                            targetSheetView.SetActiveCell(0, targetSheetView.ActiveColumnIndex)
                            targetSpreadView.ShowCell(targetSpreadView.GetActiveRowViewportIndex(), targetSpreadView.GetActiveColumnViewportIndex(), targetSheetView.ActiveRowIndex, targetSheetView.ActiveColumnIndex, FarPoint.Win.Spread.VerticalPosition.Nearest, FarPoint.Win.Spread.HorizontalPosition.Nearest)
                            Return
                        End If
                    End If

                    If activeRow = sheetView.RowCount - 1 Then
                        ' there is no sibling sheet below this one, and no child sheets for the next row, go to the next parent row
                        targetSheetView = parentSpreadView.GetSheetView()
                        While parentSpreadView IsNot Nothing AndAlso parentRowIndex + 1 >= targetSheetView.RowCount
                            ' no more rows at this level, go up one more level
                            parentRowIndex = targetSheetView.ParentRowIndex
                            parentSpreadView = parentSpreadView.Parent
                            targetSheetView = If(parentSpreadView Is Nothing, Nothing, parentSpreadView.GetSheetView())
                        End While
                        If parentSpreadView IsNot Nothing Then
                            spreadView.GetRootWorkbook().SetActiveWorkbook(parentSpreadView)
                            targetSheetView.SetActiveCell(parentRowIndex + 1, targetSheetView.ActiveColumnIndex)
                            parentSpreadView.ShowCell(parentSpreadView.GetActiveRowViewportIndex(), parentSpreadView.GetActiveColumnViewportIndex(), targetSheetView.ActiveRowIndex, targetSheetView.ActiveColumnIndex, FarPoint.Win.Spread.VerticalPosition.Nearest, FarPoint.Win.Spread.HorizontalPosition.Nearest)
                            Return
                        End If
                    End If

                Else
                    ' spreadView is top-level view (root workbook)
                    Dim children As ArrayList = spreadView.GetChildWorkbooks()
                    For Each child As FarPoint.Win.Spread.SpreadView In children
                        If child.GetSheetView().ParentRowIndex = activeRow Then
                            Dim targetSheetView As FarPoint.Win.Spread.SheetView = child.GetSheetView()
                            spreadView.GetRootWorkbook().SetActiveWorkbook(child)
                            targetSheetView.SetActiveCell(0, targetSheetView.ActiveColumnIndex)
                            child.ShowCell(child.GetActiveRowViewportIndex(), child.GetActiveColumnViewportIndex(), targetSheetView.ActiveRowIndex, targetSheetView.ActiveColumnIndex, FarPoint.Win.Spread.VerticalPosition.Nearest, FarPoint.Win.Spread.HorizontalPosition.Nearest)
                            Return
                        End If
                    Next
                End If
                ' do default action
                spreadView.GetRootWorkbook().GetActionMap().[Get]("MoveToNextRow").PerformAction(sender)
            End If
        End Sub

    End Class



    Private Sub HierarchyNavigate1_Load(sender As Object, e As EventArgs) Handles MyBase.Load
        FpSpread1.GetActionMap().Put("HierarchyMoveToPreviousRow", New HierarchyMoveToPreviousRowAction())
        FpSpread1.GetActionMap().Put("HierarchyMoveToNextRow", New HierarchyMoveToNextRowAction())
        Dim inputMap As FarPoint.Win.Spread.InputMap = FpSpread1.GetInputMap(FarPoint.Win.Spread.InputMapMode.WhenFocused)
        inputMap.Put(New FarPoint.Win.Spread.Keystroke(Keys.Up, Keys.None), "HierarchyMoveToPreviousRow")
        inputMap.Put(New FarPoint.Win.Spread.Keystroke(Keys.Down, Keys.None), "HierarchyMoveToNextRow")
        FpSpread1.DataSource = CreateDataSource()
        ExpandChildrenRecursive(FpSpread1_Sheet1)

        Dim root As FarPoint.Win.Spread.SpreadView = FpSpread1.GetRootWorkbook()
        For Each child As FarPoint.Win.Spread.SpreadView In root.GetChildWorkbooks()
            SetInputMapRecursive(inputMap, root, child)
        Next
    End Sub
End Class
